@@ -46,7 +46,7 @@
 #include "src/dbus.h"
 
 struct ap_state {
-	struct device *device;
+	struct netdev *netdev;
 	char *ssid;
 	int channel;
 	unsigned int ciphers;
@@ -55,7 +55,6 @@ struct ap_state {
 	uint8_t pmk[32];
 	struct l_queue *frame_watch_ids;
 	uint32_t start_stop_cmd_id;
-	uint32_t netdev_watch_id;
 
 	uint16_t last_aid;
 	struct l_queue *sta_states;
@@ -75,13 +74,12 @@ struct sta_state {
 	uint32_t assoc_resp_cmd_id;
 	struct ap_state *ap;
 	uint8_t *assoc_rsne;
-	size_t assoc_rsne_len;
 	struct eapol_sm *sm;
 	struct handshake_state *hs;
 };
 
 static struct l_genl_family *nl80211 = NULL;
-static uint32_t device_watch;
+static uint32_t netdev_watch;
 
 static void ap_sta_free(void *data)
 {
@@ -112,7 +110,7 @@ static void ap_frame_watch_remove(void *data, void *user_data)
 
 static void ap_reset(struct ap_state *ap)
 {
-	struct netdev *netdev = device_get_netdev(ap->device);
+	struct netdev *netdev = ap->netdev;
 
 	if (ap->pending)
 		dbus_pending_reply(&ap->pending,
@@ -128,8 +126,6 @@ static void ap_reset(struct ap_state *ap)
 	if (ap->start_stop_cmd_id)
 		l_genl_family_cancel(nl80211, ap->start_stop_cmd_id);
 
-	netdev_watch_remove(netdev, ap->netdev_watch_id);
-
 	l_queue_destroy(ap->sta_states, ap_sta_free);
 
 	if (ap->rates)
@@ -137,7 +133,7 @@ static void ap_reset(struct ap_state *ap)
 
 	ap->started = false;
 
-	l_dbus_property_changed(dbus_get_bus(), device_get_path(ap->device),
+	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(ap->netdev),
 						IWD_AP_INTERFACE, "Started");
 }
 
@@ -153,10 +149,18 @@ static void ap_free(void *data)
 static void ap_del_station(struct sta_state *sta, uint16_t reason,
 				bool disassociate)
 {
-	netdev_del_station(device_get_netdev(sta->ap->device), sta->addr,
-				reason, disassociate);
+	netdev_del_station(sta->ap->netdev, sta->addr, reason, disassociate);
 	sta->associated = false;
 	sta->rsna = false;
+
+	if (sta->sm)
+		eapol_sm_free(sta->sm);
+
+	if (sta->hs)
+		handshake_state_free(sta->hs);
+
+	sta->hs = NULL;
+	sta->sm = NULL;
 }
 
 static bool ap_sta_match_addr(const void *a, const void *b)
@@ -202,7 +206,7 @@ static void ap_new_rsna(struct sta_state *sta)
 static void ap_drop_rsna(struct sta_state *sta)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(sta->ap->device);
+	uint32_t ifindex = netdev_get_ifindex(sta->ap->netdev);
 	struct nl80211_sta_flag_update flags = {
 		.mask = (1 << NL80211_STA_FLAG_AUTHORIZED) |
 			(1 << NL80211_STA_FLAG_MFP),
@@ -233,7 +237,12 @@ static void ap_drop_rsna(struct sta_state *sta)
 		l_error("Issuing DEL_KEY failed");
 	}
 
-	handshake_state_free(sta->hs);
+	if (sta->sm)
+		eapol_sm_free(sta->sm);
+
+	if (sta->hs)
+		handshake_state_free(sta->hs);
+
 	sta->hs = NULL;
 	sta->sm = NULL;
 }
@@ -259,7 +268,7 @@ static size_t ap_build_beacon_pr_head(struct ap_state *ap,
 	struct mmpdu_header *mpdu = (void *) out_buf;
 	unsigned int len;
 	uint16_t capability = IE_BSS_CAP_ESS | IE_BSS_CAP_PRIVACY;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	uint32_t minr, maxr, count, r;
 	uint8_t *rates;
 	struct ie_tlv_builder builder;
@@ -340,7 +349,7 @@ static uint32_t ap_send_mgmt_frame(struct ap_state *ap,
 					void *user_data)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 	uint32_t id;
 	uint32_t ch_freq = scan_channel_to_freq(ap->channel, SCAN_BAND_2_4_GHZ);
 
@@ -374,6 +383,7 @@ static void ap_handshake_event(struct handshake_state *hs,
 		netdev_handshake_failed(hs, l_get_u16(event_data));
 		/* fall through */
 	case HANDSHAKE_EVENT_SETTING_KEYS_FAILED:
+		sta->sm = NULL;
 		ap_remove_sta(sta);
 	default:
 		break;
@@ -384,7 +394,7 @@ static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
 {
 	struct sta_state *sta = user_data;
 	struct ap_state *ap = sta->ap;
-	struct netdev *netdev = device_get_netdev(sta->ap->device);
+	struct netdev *netdev = sta->ap->netdev;
 	const uint8_t *own_addr = netdev_get_address(netdev);
 	struct scan_bss bss;
 	struct ie_rsn_info rsn;
@@ -402,32 +412,25 @@ static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
 	/*
 	 * TODO: This assumes the length that ap_set_rsn_info() requires. If
 	 * ap_set_rsn_info() changes then this will need to be updated.
-	 * Alternatively, sta->assoc_rsne could be used instead, but not
-	 * how it sits currently. sta->assoc_rsne only includes the actual RSN
-	 * data, not the IE type or length in the header.
 	 */
 	ie_build_rsne(&rsn, bss_rsne);
 
-	if (memcmp(bss_rsne + 2, sta->assoc_rsne, sta->assoc_rsne_len)) {
-		l_error("RSNE from association does not match");
-		goto error;
-	}
-
 	/* this handshake setup assumes PSK network */
 	sta->hs = netdev_handshake_state_new(netdev);
 
 	handshake_state_set_event_func(sta->hs, ap_handshake_event, sta);
 	handshake_state_set_ssid(sta->hs, (void *)ap->ssid, strlen(ap->ssid));
-	/* ap_rsn/own_rsn can be set equal since the check above matched */
 	handshake_state_set_ap_rsn(sta->hs, bss_rsne);
-	handshake_state_set_own_rsn(sta->hs, bss_rsne);
+	handshake_state_set_own_rsn(sta->hs, sta->assoc_rsne);
 	handshake_state_set_pmk(sta->hs, ap->pmk, 32);
 	handshake_state_set_authenticator_address(sta->hs, own_addr);
 	handshake_state_set_supplicant_address(sta->hs, sta->addr);
+	handshake_state_set_authenticator(sta->hs, true);
 
 	sta->sm = eapol_sm_new(sta->hs);
 	if (!sta->sm) {
 		handshake_state_free(sta->hs);
+		sta->hs = NULL;
 		l_error("could not create sm object");
 		goto error;
 	}
@@ -435,7 +438,7 @@ static void ap_associate_sta_cb(struct l_genl_msg *msg, void *user_data)
 	eapol_sm_set_listen_interval(sta->sm, sta->listen_interval);
 	eapol_sm_set_protocol_version(sta->sm, EAPOL_PROTOCOL_VERSION_2004);
 
-	eapol_register_authenticator(sta->sm);
+	eapol_register(sta->sm);
 
 	return;
 
@@ -446,7 +449,7 @@ error:
 static void ap_associate_sta(struct ap_state *ap, struct sta_state *sta)
 {
 	struct l_genl_msg *msg;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 	/*
 	 * This should hopefully work both with and without
 	 * NL80211_FEATURE_FULL_AP_CLIENT_STATE.
@@ -550,7 +553,7 @@ static uint32_t ap_assoc_resp(struct ap_state *ap, struct sta_state *sta,
 				enum mmpdu_reason_code status_code,
 				bool reassoc, l_genl_msg_func_t callback)
 {
-	const uint8_t *addr = device_get_address(ap->device);
+	const uint8_t *addr = netdev_get_address(ap->netdev);
 	uint8_t mpdu_buf[128];
 	struct mmpdu_header *mpdu = (void *) mpdu_buf;
 	struct mmpdu_association_response *resp;
@@ -649,7 +652,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 	struct ap_state *ap = sta->ap;
 	const char *ssid = NULL;
 	const uint8_t *rsn = NULL;
-	size_t ssid_len = 0, rsn_len = 0;
+	size_t ssid_len = 0;
 	struct l_uintset *rates = NULL;
 	struct ie_rsn_info rsn_info;
 	int err;
@@ -684,8 +687,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 				goto bad_frame;
 			}
 
-			rsn = (const uint8_t *) ie_tlv_iter_get_data(ies);
-			rsn_len = ie_tlv_iter_get_length(ies);
+			rsn = (const uint8_t *) ie_tlv_iter_get_data(ies) - 2;
 			break;
 		}
 
@@ -736,8 +738,7 @@ static void ap_assoc_reassoc(struct sta_state *sta, bool reassoc,
 	if (sta->assoc_rsne)
 		l_free(sta->assoc_rsne);
 
-	sta->assoc_rsne = l_memdup(rsn, rsn_len);
-	sta->assoc_rsne_len = rsn_len;
+	sta->assoc_rsne = l_memdup(rsn, rsn[1] + 2);
 
 	/* 802.11-2016 11.3.5.3 j) */
 	if (sta->rsna)
@@ -786,7 +787,7 @@ static void ap_assoc_req_cb(struct netdev *netdev,
 	struct sta_state *sta;
 	const uint8_t *from = hdr->address_2;
 	const struct mmpdu_association_request *req = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct ie_tlv_iter iter;
 
 	l_info("AP Association Request from %s", util_address_to_string(from));
@@ -820,7 +821,7 @@ static void ap_reassoc_req_cb(struct netdev *netdev,
 	struct sta_state *sta;
 	const uint8_t *from = hdr->address_2;
 	const struct mmpdu_reassociation_request *req = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct ie_tlv_iter iter;
 	int err;
 
@@ -877,7 +878,7 @@ static void ap_probe_req_cb(struct netdev *netdev,
 	size_t ssid_len = 0, ssid_list_len = 0, len;
 	int dsss_channel = -1;
 	struct ie_tlv_iter iter;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	bool match = false;
 	uint8_t resp[512];
 
@@ -966,7 +967,7 @@ static void ap_disassoc_cb(struct netdev *netdev,
 	struct ap_state *ap = user_data;
 	struct sta_state *sta;
 	const struct mmpdu_disassociation *disassoc = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 
 	l_info("AP Disassociation from %s, reason %i",
 		util_address_to_string(hdr->address_2),
@@ -1001,7 +1002,7 @@ static void ap_auth_reply_cb(struct l_genl_msg *msg, void *user_data)
 static void ap_auth_reply(struct ap_state *ap, const uint8_t *dest,
 				enum mmpdu_reason_code status_code)
 {
-	const uint8_t *addr = device_get_address(ap->device);
+	const uint8_t *addr = netdev_get_address(ap->netdev);
 	uint8_t mpdu_buf[64];
 	struct mmpdu_header *mpdu = (struct mmpdu_header *) mpdu_buf;
 	struct mmpdu_authentication *auth;
@@ -1036,7 +1037,7 @@ static void ap_auth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	struct ap_state *ap = user_data;
 	const struct mmpdu_authentication *auth = body;
 	const uint8_t *from = hdr->address_2;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 	struct sta_state *sta;
 
 	l_info("AP Authentication from %s", util_address_to_string(from));
@@ -1105,7 +1106,7 @@ static void ap_deauth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	struct ap_state *ap = user_data;
 	struct sta_state *sta;
 	const struct mmpdu_deauthentication *deauth = body;
-	const uint8_t *bssid = device_get_address(ap->device);
+	const uint8_t *bssid = netdev_get_address(ap->netdev);
 
 	l_info("AP Deauthentication from %s, reason %i",
 		util_address_to_string(hdr->address_2),
@@ -1125,20 +1126,6 @@ static void ap_deauth_cb(struct netdev *netdev, const struct mmpdu_header *hdr,
 	ap_sta_free(sta);
 }
 
-static void ap_netdev_notify(struct netdev *netdev,
-				enum netdev_watch_event event, void *user_data)
-{
-	struct ap_state *ap = user_data;
-
-	switch (event) {
-	case NETDEV_WATCH_EVENT_DOWN:
-		ap_reset(ap);
-		break;
-	default:
-		break;
-	}
-}
-
 static void ap_start_cb(struct l_genl_msg *msg, void *user_data)
 {
 	struct ap_state *ap = user_data;
@@ -1163,7 +1150,7 @@ static void ap_start_cb(struct l_genl_msg *msg, void *user_data)
 
 	ap->started = true;
 
-	l_dbus_property_changed(dbus_get_bus(), device_get_path(ap->device),
+	l_dbus_property_changed(dbus_get_bus(), netdev_get_path(ap->netdev),
 						IWD_AP_INTERFACE, "Started");
 }
 
@@ -1175,8 +1162,8 @@ static struct l_genl_msg *ap_build_cmd_start_ap(struct ap_state *ap)
 	size_t head_len, tail_len;
 
 	uint32_t dtim_period = 3;
-	uint32_t ifindex = device_get_ifindex(ap->device);
-	struct wiphy *wiphy = device_get_wiphy(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
+	struct wiphy *wiphy = netdev_get_wiphy(ap->netdev);
 	uint32_t hidden_ssid = NL80211_HIDDEN_SSID_NOT_IN_USE;
 	uint32_t nl_ciphers = ie_rsn_cipher_suite_to_cipher(ap->ciphers);
 	uint32_t nl_akm = CRYPTO_AKM_PSK;
@@ -1237,8 +1224,8 @@ static struct l_genl_msg *ap_build_cmd_start_ap(struct ap_state *ap)
 static int ap_start(struct ap_state *ap, const char *ssid, const char *psk,
 		struct l_dbus_message *message)
 {
-	struct netdev *netdev = device_get_netdev(ap->device);
-	struct wiphy *wiphy = device_get_wiphy(ap->device);
+	struct netdev *netdev = ap->netdev;
+	struct wiphy *wiphy = netdev_get_wiphy(netdev);
 	struct l_genl_msg *cmd;
 	const struct l_queue_entry *entry;
 	uint32_t id;
@@ -1307,8 +1294,6 @@ static int ap_start(struct ap_state *ap, const char *ssid, const char *psk,
 		goto error;
 	}
 
-	ap->netdev_watch_id = netdev_watch_add(netdev, ap_netdev_notify, ap);
-
 	ap->pending = l_dbus_message_ref(message);
 
 	return 0;
@@ -1345,7 +1330,7 @@ end:
 static struct l_genl_msg *ap_build_cmd_stop_ap(struct ap_state *ap)
 {
 	struct l_genl_msg *cmd;
-	uint32_t ifindex = device_get_ifindex(ap->device);
+	uint32_t ifindex = netdev_get_ifindex(ap->netdev);
 
 	cmd = l_genl_msg_new_sized(NL80211_CMD_STOP_AP, 16);
 	l_genl_msg_append_attr(cmd, NL80211_ATTR_IFINDEX, 4, &ifindex);
@@ -1445,34 +1430,43 @@ static void ap_destroy_interface(void *user_data)
 	ap_free(ap);
 }
 
-static void ap_add_interface(struct device *device)
+static void ap_add_interface(struct netdev *netdev)
 {
 	struct ap_state *ap;
 
 	/* just allocate/set device, Start method will complete setup */
 	ap = l_new(struct ap_state, 1);
-	ap->device = device;
+	ap->netdev = netdev;
 
 	/* setup ap dbus interface */
 	l_dbus_object_add_interface(dbus_get_bus(),
-			device_get_path(device), IWD_AP_INTERFACE, ap);
+			netdev_get_path(netdev), IWD_AP_INTERFACE, ap);
 }
 
-static void ap_remove_interface(struct device *device)
+static void ap_remove_interface(struct netdev *netdev)
 {
 	l_dbus_object_remove_interface(dbus_get_bus(),
-			device_get_path(device), IWD_AP_INTERFACE);
+			netdev_get_path(netdev), IWD_AP_INTERFACE);
 }
 
-static void ap_device_event(struct device *device, enum device_event event,
-								void *userdata)
+static void ap_netdev_watch(struct netdev *netdev,
+				enum netdev_watch_event event, void *userdata)
 {
+	struct device *device = netdev_get_device(netdev);
+
+	if (!device)
+		return;
+
 	switch (event) {
-	case DEVICE_EVENT_MODE_CHANGED:
-		if (device_get_mode(device) == DEVICE_MODE_AP)
-			ap_add_interface(device);
-		else
-			ap_remove_interface(device);
+	case NETDEV_WATCH_EVENT_UP:
+	case NETDEV_WATCH_EVENT_NEW:
+		if (netdev_get_iftype(netdev) == NETDEV_IFTYPE_AP)
+			ap_add_interface(netdev);
+		break;
+	case NETDEV_WATCH_EVENT_DOWN:
+	case NETDEV_WATCH_EVENT_DEL:
+		ap_remove_interface(netdev);
+		break;
 	default:
 		break;
 	}
@@ -1480,10 +1474,7 @@ static void ap_device_event(struct device *device, enum device_event event,
 
 bool ap_init(struct l_genl_family *in)
 {
-	device_watch = device_watch_add(ap_device_event, NULL, NULL);
-	if (!device_watch)
-		return false;
-
+	netdev_watch = netdev_watch_add(ap_netdev_watch, NULL, NULL);
 	nl80211 = in;
 
 	return l_dbus_register_interface(dbus_get_bus(), IWD_AP_INTERFACE,
@@ -1496,7 +1487,6 @@ bool ap_init(struct l_genl_family *in)
 
 void ap_exit(void)
 {
-	device_watch_remove(device_watch);
-
+	netdev_watch_remove(netdev_watch);
 	l_dbus_unregister_interface(dbus_get_bus(), IWD_AP_INTERFACE);
 }
